/* (C) Pavel Surynek, 2009 - 2010  */
/* pavel.surynek@mff.cuni.cz       */
/* graph.h                         */

/*
  Graph representation.
 */


#ifndef __GRAPH_H__
#define __GRAPH_H__

class Vertex;
class IVertex;
class Literal;
class Clause;
class Formula;
class VMatrix;


class VMatrix
{
public:
	int init(int L, int max_visits);
	void print(void);

public:
	int m_L, m_max_visits;
	int **m_cPaths;
};


class Edge
{
public:
	int m_weight;
};


class Neighbor
{
public:
	Edge *m_edge;
	Vertex *m_vertex;
};


class Vertex
{
public:
	int m_N, m_cN;
	Neighbor *m_neighbors;
};


class Graph
{
public:
	int init(int V, int E, int *cN);
	void add_edge(int u, int v, int weight = 0);

	void print(void) const;

public:
	int m_V;
	Vertex *m_vertices;

	int m_E, m_cE;
	Edge *m_edges;

	int m_N, m_cN;
	Neighbor *m_neighbors;
};


class IEdge
{
public:
	int m_weight;
};


class INeighborNode
{
public:
	IEdge *m_edge;
	IVertex *m_vertex;
	INeighborNode *m_next;
};


class IVertex
{
public:
	int m_id;
	Literal *m_literal;

	int m_cluster;
	int m_layer;
	int m_visits;

	int m_cN;
	INeighborNode *m_first_nnode;
	INeighborNode **m_last_nnode;

// path search specific members
	int m_distance;
	VMatrix m_vmatrix;
};


class IVertexChunk
{
public:
	IVertex *m_vertices;
	IVertexChunk *m_next;
};


class IEdgeChunk
{
public:
	IEdge *m_edges;
	IEdgeChunk *m_next;
};


class INeighborChunk
{
public:
	INeighborNode *m_nnodes;
	INeighborChunk *m_next;
};


class VONode
{
public:
	IVertex *m_vertex;
	VONode *m_next;	
};


class IGraph
{
public:
	int init(Formula *formula, int V_chunk, int E_chunk, int N_chunk);
	int add_vertex(int cV = 1);
	int add_edge(int u, int v, int weight = 0);
	int increase_edge_weight(int u, int v, int dweight = 1);

	IVertex *get_vertex(int id);
        IEdge *get_edge(int u, int v);

	void print(void) const;
	void export_GR(bool show_literals = false, bool exp_clusters = false, bool exp_layers = false, bool exp_visits = false) const;

	void get_label(char *label, IVertex *vertex, bool exp_literals = false) const;
	void get_data(char *data, IVertex *vertex, bool exp_clusters = false, bool exp_layers = false, bool exp_visits = false) const;

private:
	int alloc_vertices(int cV);
	int alloc_edge(IEdge **edge);
	int alloc_neighbor(INeighborNode **nnode);

public:
	Formula *m_formula;
	int m_last_id;

	int m_V, m_cV, m_ccV;
	int m_V_chunk;
	IVertexChunk *m_first_vchunk;
	IVertexChunk **m_last_vchunk;

	int m_E, m_cE, m_ccE;
	int m_E_chunk;
	IEdgeChunk *m_first_echunk;
	IEdgeChunk **m_last_echunk;

	int m_N, m_cN, m_ccN;
	int m_N_chunk;
	INeighborChunk *m_first_nchunk;
	INeighborChunk **m_last_nchunk;

	int m_cVO;
	VONode *m_vo_nodes;
};


class Cluster
{
public:
	int m_cV;
	IVertex **m_vertices;
};


class Clustering
{
public:
	int init(const Formula *formula, Clause **path, int length, IGraph *graph);

	void print(void) const;
	void print_distances(void) const;
	void print_visits(void) const;
	void print_visit_matrices(void) const;

public:
	const Formula *m_formula;
	IGraph *m_graph;

	int m_cC;
	Cluster *m_clusters;
};


class Layer
{
public:
	int m_V, m_cV;
	IVertex **m_vertices;
};


class Layering
{
public:
	int init(const Formula *formula, Clause **path, int length, IGraph *graph);
	int init_visit_matrices(int L, int max_visits);

	void print(void) const;
	void print_visits(void) const;
public:
	const Formula *m_formula;
	IGraph *m_graph;

	int m_L, m_cL;
	Layer *m_layers;
};



#endif
